import * as React from 'react';
import {
  newInstance,
  BrowserJsPlumbInstance,
  BrowserJsPlumbDefaults,
} from '@jsplumb/browser-ui';
import { useEffect, useRef, useState } from 'react';
import {
  NODE_ID_ATTR,
  NODE_PARENT_ATTR,
  NODE_ID_LEN,
  DEFAULT_TREE_OPTION,
  DEFAULT_CONNECTION_OPTION,
  DEFAULT_NODE_STYLE,
  HIGHLIGHT_CONNECTION_OPTION,
} from './constant';
import style from './index.module.css';

// default template element to render node
const DefaultNodeTemplate = (props: TreeNode) => {
  if (props.data) {
    return (
      <div
        className={style['connect-tree-node-content']}
        style={DEFAULT_NODE_STYLE}
      >
        {props.data}
      </div>
    );
  } else {
    return <div></div>;
  }
};

// define interface and type-------------------------------------
export interface TreeNode {
  data: any;
  children?: TreeNode[];
  [NODE_ID_ATTR]?: string;
  [NODE_PARENT_ATTR]?: TreeNode;
}

export interface NodesRelation {
  parent: TreeNode | null;
  child: TreeNode;
}

export interface tranverseTreeByLayers {
  (root: TreeNode): {
    nodesByLayers: TreeNode[][];
    nodesRelations: NodesRelation[];
  };
}

export interface ConnectTreeRef {
  [nodeRef: string]: HTMLDivElement | undefined;
}

export interface ConnectTreeParams {
  dataSource: TreeNode;
  connectionOption?: BrowserJsPlumbDefaults;
  canHighlight?: boolean;
  style?: React.CSSProperties;
  highlightOption?: BrowserJsPlumbDefaults;
  renderChildNode?: (props: any) => JSX.Element;
}

// some util function ------------------------------------------------------
// generate random tree data for test
export const randomGeneratesTree = (
  maxLayer: number = 1,
  maxNodeNum: number = 1,
  maxChildrenPerNode: number = 1
): TreeNode => {
  const root = {
    data: 'layer 0-0',
    children: [],
  };
  let totalNodeNum = 1;
  const dfs = (root: TreeNode, currentLayer: number) => {
    if (root && currentLayer <= maxLayer && totalNodeNum <= maxNodeNum) {
      let childNum = 0;
      // random to generate a child
      let goNextRatio = Math.random();
      while (
        goNextRatio < 0.8 &&
        childNum < maxChildrenPerNode &&
        totalNodeNum < maxNodeNum
      ) {
        // create a new node
        const child: TreeNode = {
          data: `layer ${currentLayer} - ${childNum}`,
          children: [],
        };
        totalNodeNum++;
        // dfs the child
        dfs(child, currentLayer + 1);
        // add child to current node
        if (root.children) {
          root.children.push(child);
        } else {
          root.children = [child];
        }
        childNum++;
        goNextRatio = Math.random();
      }
    }
  };

  dfs(root, 1);
  return root;
};

// generate a random and unique string
const genRandomUniqueID = (length: number): string => {
  return Number(
    Math.random()
      .toString()
      .substr(3, length) + Date.now()
  ).toString(36);
};

// tranverse tree by layers
export const tranverseTreeByLayers: tranverseTreeByLayers = (
  root: TreeNode
) => {
  const queue: TreeNode[] = [];
  // treeNode place by layers: layer0[treenode ...] layer1[treenode]
  const nodesByLayers: TreeNode[][] = [];
  const nodesRelations: NodesRelation[] = [];

  // start tranversing tree
  queue.push(root);
  while (queue.length > 0) {
    const numInThisLayer = queue.length;
    const nodesInThisLayer: TreeNode[] = [];
    for (let i = 0; i < numInThisLayer; i++) {
      const node = queue.shift();
      if (node && node.data) {
        // add the node unique id
        node[NODE_ID_ATTR] = genRandomUniqueID(NODE_ID_LEN);
        // add node to this layer
        nodesInThisLayer.push(node);
        if (node.children) {
          // add children to queue (next layer)
          node.children.forEach(child => {
            queue.push(child);
            child[NODE_PARENT_ATTR] = node;
            // save the relation for connection
            nodesRelations.push({
              parent: node,
              child: child,
            });
          });
        }
      }
    }
    // complete one layer
    nodesInThisLayer && nodesByLayers.push(nodesInThisLayer);
  }

  return {
    nodesByLayers,
    nodesRelations,
  };
};

// the main component ----------------------------------------------------
/* TODO:
  1. batch绘制connection（节点出现1000+时，绘制线条有明显卡顿）
     --> 1000节点，不用batch时间：1500ms~6000ms；用batch时间：500~1500ms
  2. 测试遍历时间，考虑是否通过useMemo
     --> 遍历节点在1000000+才会消耗10+ms，10000节点基本在1ms。感觉没有必要useMemo?
  3. 输入数据不变时，如何避免重复渲染（node and connection）
     --> 目前datasource的引用不变不会重绘connection；使用React.memo包裹时props中引用不变不会渲染
  4. 渲染节点相当多时，考虑分片更新？
  5. 收起、展开节点？
  6. 虚拟滚动？
  7. 超过容器的线不应该绘制，容器用relative定位时scroll会错位
*/
const ConnectTree = (props: ConnectTreeParams) => {
  // props setting
  const root = props.dataSource || {};
  const connectionOption = props.connectionOption || DEFAULT_CONNECTION_OPTION;
  const highLightOption = props.highlightOption || HIGHLIGHT_CONNECTION_OPTION;
  const renderChildNode = props.renderChildNode || DefaultNodeTemplate;
  // tree option
  const canHighlight =
    props.canHighlight === false ? false : DEFAULT_TREE_OPTION.canHighlight;
  const treeStyle = props.style || {};
  // hooks init
  const domRefs = useRef<ConnectTreeRef>({});
  const plumbInstanceRef = useRef<BrowserJsPlumbInstance | null>(null);
  const [highlightRelations, setHighlightRelations] = useState<NodesRelation[]>(
    []
  );

  const updateConnection = (
    realtion: NodesRelation[],
    highlightRelations: NodesRelation[],
    currentInstance: BrowserJsPlumbInstance | null,
    nodeRefs: React.MutableRefObject<ConnectTreeRef>,
    connectionOption: object,
    highLightOption: object
  ) => {
    // clear the previos connections
    currentInstance && currentInstance.deleteEveryConnection();
    // batch update connecntion for better performance
    currentInstance &&
      currentInstance.batch(() => {
        realtion.forEach(({ parent, child }) => {
          // build new connections
          if (parent && parent[NODE_ID_ATTR] && child && child[NODE_ID_ATTR]) {
            currentInstance.connect(
              {
                source: nodeRefs.current[parent[NODE_ID_ATTR] as string],
                target: nodeRefs.current[child[NODE_ID_ATTR] as string],
              },
              connectionOption
            );
          }
        });
        highlightRelations.forEach(({ parent, child }) => {
          // build new connections
          if (parent && parent[NODE_ID_ATTR] && child && child[NODE_ID_ATTR]) {
            currentInstance.connect(
              {
                source: nodeRefs.current[parent[NODE_ID_ATTR] as string],
                target: nodeRefs.current[child[NODE_ID_ATTR] as string],
              },
              {
                ...connectionOption,
                ...highLightOption,
              }
            );
          }
        });
      });
  };

  const handleClickNode: React.MouseEventHandler<HTMLDivElement> = e => {
    if (!canHighlight) return;

    const node =
      e.currentTarget &&
      ((e.currentTarget as any)._treeNode as TreeNode);
    const path: NodesRelation[] = [];

    let curr = node;
    // find previous nodes path
    while (curr && curr[NODE_PARENT_ATTR]) {
      const parent = curr[NODE_PARENT_ATTR] as TreeNode;
      path.push({
        parent: parent,
        child: curr,
      });
      curr = parent;
    }
    // find next nodes path
    const dfs = (node: TreeNode) => {
      if (!node || !node.children) {
        return;
      } else {
        node.children.forEach(child => {
          path.push({
            parent: node,
            child: child,
          });
          dfs(child);
        });
      }
    };
    dfs(node);
    setHighlightRelations(path);
  };

  // repaint all connection when window reize
  const handleResizeWindow = () => {
    plumbInstanceRef.current?.repaintEverything();
  };

  // init plumb instance and event listener
  useEffect(() => {
    // 这里没有更新作用的状态改为useRef
    // const instance = newInstance({ ...connectionOption, container: domRefs.current['connect-tree-container'] });
    // setPlumbInstance(instance);

    plumbInstanceRef.current = newInstance({
      ...connectionOption,
      container: domRefs.current['connect-tree-container']!,
    });

    // add listener for window-resize event
    window.addEventListener('resize', handleResizeWindow);
    return () => {
      plumbInstanceRef.current?.destroy();
      window.removeEventListener('resize', handleResizeWindow);
    };
  }, []);

  // start tranverse the tree from the root
  const { nodesByLayers, nodesRelations } = tranverseTreeByLayers(root);

  // update connection when props(source/template) or state(plumb) change
  useEffect(() => {
    const _startTime = new Date().getTime();
    //  start tranverse the tree from the root
    updateConnection(
      nodesRelations,
      highlightRelations,
      plumbInstanceRef.current,
      domRefs,
      connectionOption,
      highLightOption
    );
    const _endTime = new Date().getTime();
    console.log(`connection update time: ${_endTime - _startTime}`);
  }, [
    root,
    renderChildNode,
    highlightRelations,
    connectionOption,
    highLightOption,
  ]);

  // test
  useEffect(() => {
    console.log('tree update');
  });

  return (
    <div
      className={style['connect-tree']}
      style={treeStyle}
      onScroll={handleResizeWindow}
      ref={el => {
        if (el) {
          domRefs.current['connect-tree-container'] = el;
        }
      }}
    >
      {nodesByLayers.map((nodesInlayer, layer) => {
        // render nodes by layers
        return (
          <div
            className={style['connect-tree-layer']}
            key={layer}
          >
            {nodesInlayer.map((node, index) => {
              const node_id = node[NODE_ID_ATTR];
              return (
                <div
                  className={style["connect-tree-node"]}
                  key={node_id || index}
                  // give every node an unique ref
                  ref={(el: any) => {
                    if (node_id && el) {
                      // 把node存到dom节点便于点击查找对应数据里不知是否合理
                      el._treeNode = node;
                      domRefs.current[node_id] = el;
                    }
                  }}
                  // 避免每个节点都创建箭头函数
                  onClick={handleClickNode}
                >
                  {/* render node */}
                  {renderChildNode(node)}
                </div>
              );
            })}
          </div>
        );
      })}
    </div>
  );
};

export default ConnectTree;
